//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>
//---------------------------------------------------------------------------
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDraw.h"
//---------------------------------------------------------------------------
TDirectDraw*	DirectDraw = NULL;
//---------------------------------------------------------------------------
static inline TDirectDraw *ValidCtrCheck()
{
	return new TDirectDraw(NULL);
}
//---------------------------------------------------------------------------
namespace Tdirectdraw
{
	void __fastcall Register()
	{
		TComponentClass classes[1] = {__classid(TDirectDraw)};
		RegisterComponents("DirectDraw", classes, 0);
	}
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::TDirectDraw(TComponent* Owner)
	: TComponent(Owner)
{
// create things used for debugging
fErrorIndent = 0;
fErrorValue = 0;
fErrorCount = 0;
fErrorStrings = new TStringList();
fErrorStrings->Sorted = false;
fErrorStrings->Duplicates = dupAccept;
fDebugging = false;

// initialize some values
fDirectDraw = NULL;
fPrimarySurface = NULL;
fBackBuffer = NULL;

fBackBufferDC = NULL;
fPrimarySurfaceDC = NULL;
fWidth = 800;
fHeight = 600;
fBitsPerPixel = 16;
fBackBufferCount = 1;

fMouseCursorBkgnd = new Graphics::TBitmap();
fMouseCursorBkgnd->Width = GetSystemMetrics(SM_CXCURSOR);
fMouseCursorBkgnd->Height = GetSystemMetrics(SM_CYCURSOR);
fMouseFirstTimeDraw = true;

// if we have created our DirectDraw object...
if (CreateDirectDraw())
	{
	// make a note of a successful result
//	Log(LOG_NOTE,"TDirectDraw::Constructor()","the DirectDraw object has been created successfully");

	// try setting exclusive mode
    if (SetExclusiveMode())
    	{
	    // make a note of a successful result
//		Log(LOG_NOTE,"TDirectDraw::Constructor()","exclusive mode has been set successfully");
        }

    // create required display surfaces
    if (CreateSurfaces())
    	{
	    // make a note of a successful result
//		Log(LOG_NOTE,"TDirectDraw::Constructor()","surfaces have been created successfully");
        }
    }
// otherwise display an error messsage
else
    ShowMessage("ERROR>> TDirectDraw::Constructor() >> Could not create the DirectDraw object");

//Log(LOG_OUT,"TDirectDraw::Constructor()",">>> The Direct Draw object has been created successfully <<<");
fDebugging = false;
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::~TDirectDraw()
{
fDebugging = false;
//Log(LOG_IN,"TDirectDraw::Destructor()","");

// if we previously created our DirectDraw object...
if (fDirectDraw != NULL)
	{
	// make a note of a successful result
//	Log(LOG_NOTE,"TDirectDraw::Destructor()","DirectDraw object was created, time to release");

	// make sure the display is pointing to the GDI surface
    FlipToGDI();
     
    // release the display surfaces
    ReleaseSurfaces();

    // restore the original video mode
    RestoreDisplayMode();

	// restore normal mode
    SetNormalMode();

    // release the directdraw object
    ReleaseDirectDraw();
    }
// otherwise display an error messsage
else
    ShowMessage("ERROR>> TDirectDraw::Constructor() >> Could not release the DirectDraw object");

// release memory used by mouse cursor routines
if (fMouseCursorBkgnd != NULL) delete fMouseCursorBkgnd;

// release things used for debugging
if (fErrorStrings != NULL) { fErrorStrings->Clear(); delete fErrorStrings; }
}
//---------------------------------------------------------------------------
// Func: CreateDirectDraw()
// Desc: simply creates a single instance of a IDirectDraw object
bool	__fastcall	TDirectDraw::CreateDirectDraw()
{
fErrorValue = DirectDrawCreate(NULL, &fDirectDraw, NULL);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateDirectDraw()","DirectDraw object could not be created");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseDirectDraw()
{
if (fDirectDraw == NULL) return false;
// simply call the Release function & use result
fErrorValue = fDirectDraw->Release();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseDirectDraw()","DirectDraw object could not be released");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetExclusiveMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetExclusiveMode()","exclusive mode could not be set");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetNormalMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_NORMAL );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetNormalMode()","normal mode could not be set");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::RestoreDisplayMode()
{
if (fDirectDraw == NULL) return false;
// simply call the RestoreDisplayMode function & use result
fErrorValue = fDirectDraw->RestoreDisplayMode();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::RestoreVideoMode()","video mode could not be restored");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::CreateSurfaces()
{
if (fDirectDraw == NULL) return false;

// create the primary surface if it's not already created
fErrorValue = DD_OK;
if (fPrimarySurface == NULL)
	{
	// define the properties of the primary surface to be created
	DDSURFACEDESC	ddSurfaceDesc;
	memset(&ddSurfaceDesc,0,sizeof(ddSurfaceDesc));

	ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);
	ddSurfaceDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	ddSurfaceDesc.dwBackBufferCount = fBackBufferCount;

	// create the primary surface
	fErrorValue = fDirectDraw->CreateSurface(&ddSurfaceDesc,&fPrimarySurface,NULL);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not create the primary surface");
    else
    	{
        // note a successful result
//        Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","the primary surface has been created");

        // only retrieve back buffer if any exist
        if (fBackBufferCount>0)
        	{
			// setup the surface caps ready to retrieve the back buffer
			DDSCAPS			ddsCaps;
			ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

	        // retrieve a pointer to the 1st back buffer in the list (i theorize)
			fErrorValue = fPrimarySurface->GetAttachedSurface(&ddsCaps,&fBackBuffer);
			if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not retrieve back buffer");
//		    else Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","back buffer has been retrieved");
            }
        }
	}
//else
//	Log(LOG_WARNING,"TDirectDraw::CreateSurfaces()","primary surface has already been created");

return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseSurfaces()
{
if (fDirectDraw == NULL) return false;
/*
NOTE:	the back buffer cannot be released directly - all associated back
		buffers are released when you release the primary surface			*/

// get rid of our primary surface & any attached back buffers if they exist
fErrorValue = DD_OK;
if (fPrimarySurface != NULL)
	{
    fErrorValue = fPrimarySurface->Release();
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseSurfaces()","could not release primary surface");
	else
    	{
        fPrimarySurface = NULL;
    	fBackBuffer = NULL;
        }
    }
//else
//	Log(LOG_WARNING,"TDirectDraw::ReleaseSurfaces()","Primary surface was not allocated, thus no buffers were released");
return (fErrorValue != DD_OK);
}
//---------------------------------------------------------------------------
HDC		__fastcall 	TDirectDraw::GetBackBufferDC()
{
// if the DC is already retrieved, skip out
if (fBackBufferDC != NULL) return NULL;
if (fPrimarySurfaceDC != NULL) return NULL;

HDC hdc;
fErrorValue = fBackBuffer->GetDC(&hdc);
//int iter = 0;
//while ((fErrorValue == DDERR_WASSTILLDRAWING) && (++iter<1000000)) fErrorValue = fBackBuffer->GetDC(&hdc);
if (fErrorValue == DDERR_SURFACELOST)
   	{
    fPrimarySurface->Restore();
    fErrorValue = fBackBuffer->GetDC(&hdc);
    }
if (fErrorValue != DD_OK)
	{
    fBackBufferDC = NULL;
    Error(fErrorValue,"TDirectDraw::GetBackBufferDC()","Could not retrieve a DC for the back buffer");
    }
else
	{
    fBackBufferDC = hdc;
//    Log(LOG_NOTE,"TDirectDraw::GetBackBufferDC()","a DC for the back buffer was retrieved");
    }

return fBackBufferDC;
}
//---------------------------------------------------------------------------
HDC		__fastcall 	TDirectDraw::GetPrimarySurfaceDC()
{
// if the DC is already retrieved, skip out
if (fBackBufferDC != NULL) return NULL;
if (fPrimarySurfaceDC != NULL) return NULL;

HDC hdc;
fErrorValue = fPrimarySurface->GetDC(&hdc);
//int iter = 0;
//while ((fErrorValue == DDERR_WASSTILLDRAWING) && (++iter<1000000)) fErrorValue = fPrimarySurface->GetDC(&hdc);
if (fErrorValue == DDERR_SURFACELOST)
   	{
    fPrimarySurface->Restore();
    fErrorValue = fPrimarySurface->GetDC(&hdc);
    }
if (fErrorValue != DD_OK)
	{
    fPrimarySurfaceDC = NULL;
    Error(fErrorValue,"TDirectDraw::GetPrimarySurfaceDC()","Could not retrieve a DC for the primary surface");
    }
else
	{
    fPrimarySurfaceDC = hdc;
//    Log(LOG_NOTE,"TDirectDraw::GetPrimarySurfaceDC()","a DC for the primary surface was retrieved");
    }

return fPrimarySurfaceDC;
}
//---------------------------------------------------------------------------
bool	__fastcall TDirectDraw::ReleaseBackBufferDC()
{
if (fPrimarySurfaceDC != NULL) return NULL;
if (fBackBufferDC == NULL) return false;

fErrorValue = fBackBuffer->ReleaseDC(fBackBufferDC);
if (fErrorValue != DD_OK)
	{
    Error(fErrorValue,"TDirectDraw::ReleaseBackBufferDC()","Could not release the back buffer DC"); // note: should _never_ happen
    return false;
    }
else
	{
	fBackBufferDC = NULL;
//    Log(LOG_NOTE,"TDirectDraw::ReleaseBackBufferDC()","the back buffer DC has been released");
    }
return true;
}
//---------------------------------------------------------------------------
bool	__fastcall TDirectDraw::ReleasePrimarySurfaceDC()
{
if (fPrimarySurfaceDC == NULL) return false;

fErrorValue = fPrimarySurface->ReleaseDC(fPrimarySurfaceDC);
if (fErrorValue != DD_OK)
	{
    Error(fErrorValue,"TDirectDraw::ReleasePrimarySurfaceDC()","Could not release the primary surface DC"); // note: should _never_ happen
    return false;
    }
else
	{
	fPrimarySurfaceDC = NULL;
//    Log(LOG_NOTE,"TDirectDraw::ReleasePrimarySurfaceDC()","the primary surface DC has been released");
    }
return true;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::Flip(bool wait)
{
if (fDirectDraw == NULL) return false;

//
// NOTE: only surfaces with DDSCAPS_FLIP and DDSCAPS_FRONTBUFFER can be flipped
//		 ie. the primary surface
//

fErrorValue = DD_OK + 1;
if (fPrimarySurface != NULL)
   	{
    int flags = 0;
    if (wait) flags |= DDFLIP_WAIT;
    fErrorValue = fPrimarySurface->Flip(NULL,flags);
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::Flip()","the surface could not be flipped");
//    else Log(LOG_NOTE,"TDirectDraw::Flip()","the surface has been flipped successfully");
//    for (int adelay=0;adelay<250000000;adelay++) ;
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipToGDI()
{
if (fDirectDraw == NULL) return false;

// simply call the FlipToGDISurface function & use result
fErrorValue = fDirectDraw->FlipToGDISurface();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipToGDI()","couldn't flip to gdi surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipPrimaryToBack(bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = 0;
r.top = 0;
r.right = fWidth-1;
r.bottom = fHeight-1;
int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fBackBuffer->BltFast(0,0,fPrimarySurface,NULL,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipGDOToBack()","couldn't flip primary to back buffer surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipBackToPrimary(bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = 0;
r.top = 0;
r.right = fWidth-1;
r.bottom = fHeight-1;
int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fPrimarySurface->BltFast(0,0,fBackBuffer,NULL,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipBackToPrimary()","couldn't flip back buffer to primary surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipPrimaryRectToBack(int left,int top,int width,int height, bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = left;
r.top = top;
r.right = left + width;
r.bottom = top + height;

if (r.left<0) r.left = 0;
if (r.top<0) r.top = 0;
if (r.right>fWidth) r.right = fWidth;
if (r.bottom>fHeight) r.bottom = fHeight;

int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fBackBuffer->BltFast(r.left,r.top,fPrimarySurface,&r,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipPrimaryRectToBack()","couldn't flip primary rect to back buffer");
if (fErrorValue == DDERR_INVALIDRECT) Error(fErrorValue,"TDirectDraw::FlipPrimaryRectToBack()","rectangle requested is invalid");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipBackRectToPrimary(int left,int top,int width,int height, bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = left;
r.top = top;
r.right = left + width;
r.bottom = top + height;

if (r.left<0) r.left = 0;
if (r.top<0) r.top = 0;
if (r.right>fWidth) r.right = fWidth;
if (r.bottom>fHeight) r.bottom = fHeight;

int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fPrimarySurface->BltFast(r.left,r.top,fBackBuffer,&r,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipBackRectToPrimary()","couldn't flip back buffer rect to primary surface");
if (fErrorValue == DDERR_INVALIDRECT) Error(fErrorValue,"TDirectDraw::FlipBackRectToPrimary()","rectangle requested is invalid");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::DrawMouseCursor(HDC hdc)
{
if (fDirectDraw == NULL) return false;
if (hdc == NULL) return false;

// do some mouse funny stuff....
POINT	mouse_pt;
if (GetCursorPos( &mouse_pt ) != 0)
	{
	fMouseCursorPos.x = mouse_pt.x;
	fMouseCursorPos.y = mouse_pt.y;
	}

// cursor drawing steps:
// (1) if it's not the first time the mouse cursor has been drawn,
//     redraw whats underneath the cursor (this stops the mouse being shadowed when it's
//     first-time drawn)
// (2) save what is currently underneath the mouse cursor
// (3) draw the mouse cursor

// draw the mouse icon
/*
TCanvas* canvas = new TCanvas();
try
	{
	canvas->Handle = hdc;

	// (1) only when there is actually something underneath the cursor
    TRect srce_r;
    TRect dest_r;
    if (!fMouseFirstTimeDraw)
    	{
	    srce_r.Left = 0;
	    srce_r.Top = 0;
	    srce_r.Right = fMouseCursorBkgnd->Width;
	    srce_r.Bottom = fMouseCursorBkgnd->Height;
	    dest_r.Left = fMouseCursorLastPos.x;
	    dest_r.Top = fMouseCursorLastPos.y;
	    dest_r.Right = dest_r.Left + fMouseCursorBkgnd->Width;
	    dest_r.Bottom = dest_r.Top + fMouseCursorBkgnd->Height;
	    // restore what was underneath the cursor the last time it was painted
	    canvas->CopyMode = cmSrcCopy;
		canvas->CopyRect( dest_r, fMouseCursorBkgnd->Canvas, srce_r );
        }

	// save what is currently underneath the mouse cursor
    srce_r.Left = fMouseCursorPos.x;
    srce_r.Top = fMouseCursorPos.y;
    srce_r.Right = srce_r.Left + fMouseCursorBkgnd->Width;
    srce_r.Bottom = srce_r.Top + fMouseCursorBkgnd->Height;
    dest_r.Left = 0;
    dest_r.Top = 0;
    dest_r.Right = fMouseCursorBkgnd->Width;
    dest_r.Bottom = fMouseCursorBkgnd->Height;
    canvas->CopyMode = cmSrcCopy;
    fMouseCursorBkgnd->Canvas->CopyRect( dest_r, canvas, srce_r );

    // draw the mouse cursor onto our fMouseCursorBkgnd
	DrawIconEx( canvas->Handle, srce_r.Left, srce_r.Top, Screen->Cursors[Screen->Cursor],0,0,0,NULL,DI_NORMAL);
	}
catch ( ... ) {}
// clean up
delete canvas;
*/

//DrawIconEx( hdc, fMouseCursorPos.x, fMouseCursorPos.y, Screen->Cursors[Screen->Cursor],0,0,0,NULL,DI_NORMAL);
return true;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutImage(HDC hdc,int left,int top,int width,int height,TImage* image,long copymode)
{
if ((hdc == NULL) || (image == NULL)) return false;
TCanvas* canvas = new TCanvas();
if (canvas == NULL) return false;

bool result;
TRect r;
r.Top = top;
r.Left = left;
r.Right = left + width;
r.Bottom = top + height;
canvas->Handle = hdc;
try
	{
    canvas->CopyMode = copymode;
    canvas->StretchDraw(r,image->Picture->Graphic);
    result = true;
    }
catch ( ... )
	{
    result = false;
    }
canvas->Handle = NULL;
delete canvas;
return result;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutBackBufferImage(int left,int top,int width,int height,TImage* image,long copymode)
{
HDC hdc = GetBackBufferDC();
if (hdc != NULL)
	{
    PutImage(hdc,left,top,width,height,image,copymode);
    ReleaseBackBufferDC();
    return true;
    }
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutPrimarySurfaceImage(int left,int top,int width,int height,TImage* image,long copymode)
{
HDC hdc = GetPrimarySurfaceDC();
if (hdc != NULL)
	{
    PutImage(hdc,left,top,width,height,image,copymode);
    ReleasePrimarySurfaceDC();
    return true;
    }
return false;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetImage(HDC hdc,int left,int top,int width,int height)
{
if (hdc == NULL) return NULL;

TImage* image = new TImage(this);
if (image == NULL) return NULL;

TCanvas* canvas = new TCanvas();
if (canvas == NULL) { delete image; return NULL; }

bool result;
try
	{
	canvas->Handle = hdc;

    TRect srce_r;
    srce_r.Left = left;
    srce_r.Top = top;
    srce_r.Right = left + width;
    srce_r.Bottom = top + height;

    image->AutoSize = true;
    image->BoundsRect = srce_r;
    image->Canvas->CopyMode = cmSrcCopy;
    image->Canvas->CopyRect( srce_r, canvas, srce_r );

    result = true;
    }
catch ( ... )
	{
    result = false;
    }

canvas->Handle = NULL;
delete canvas;

// only return the image if everything worked smoothly
if (result) return image;

delete image;
return NULL;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetBackBufferImage(int left,int top,int width,int height)
{
HDC hdc = GetBackBufferDC();
if (hdc != NULL)
	{
    TImage* image = GetImage(hdc,left,top,width,height);
    ReleaseBackBufferDC();
    return image;
    }
return NULL;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetPrimarySurfaceImage(int left,int top,int width,int height)
{
HDC hdc = GetPrimarySurfaceDC();
if (hdc != NULL)
	{
    TImage* image = GetImage(hdc,left,top,width,height);
    ReleasePrimarySurfaceDC();
    return image;
    }
return NULL;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetDisplayMode(int width,int height,int bitsperpixel)
{
if (fDirectDraw == NULL) return false;

// since the display mode can only be changed when in "Exclusive Mode"...
fErrorValue = DD_OK;

// make sure to release the surfaces & change the mode if and only if required
if ((width != fWidth) || (height != fHeight) || (bitsperpixel != fBitsPerPixel))
	{
    // tell DirectDraw to change the display mode to that required
	fErrorValue = fDirectDraw->SetDisplayMode(width,height,bitsperpixel);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetDisplayMode()","could not set display mode to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
	else
		{
		fWidth = width;
		fHeight = height;
		fBitsPerPixel = bitsperpixel;

        // re-create the surfaces now the resolution has changed
        ReleaseSurfaces();
        CreateSurfaces();
        }
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg )
{
if (fErrorCount>5000) return;
if (!fDebugging) return;
TStringList*	new_strings = new TStringList();
if (new_strings == NULL) return;	// blah

// keep track of overall memory usage
//if (overall_FirstTimeIn)
//	{
//    overall_FirstTimeIn = false;
//    InMemoryStatus( &overall_in_mstat );
//    }
//if ((msgclass == LOG_OUT) || (msgclass == LOG_IN))
//	{
//	OutMemoryStatus( &overall_out_mstat );
//	if (!SameMemoryStatus( &overall_in_mstat, &overall_out_mstat ))
//		DumpMemoryStatus( "OVERALL MEMORY USAGE", &overall_in_mstat, &overall_out_mstat );
//    }

// simple patch
System::AnsiString	indent = "";
int indenture;
for (indenture=0;indenture<fErrorIndent;indenture++) indent = indent + "    ";

switch (msgclass)
	{
    case	LOG_IN:
    		if (fErrorIndent == 0) new_strings->Add(indent + "-----------------------------------------------------------------------------");
            else new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc);
            fErrorIndent ++;
            break;

    case	LOG_OUT:
			fErrorIndent --;
			indent = "";
			for (indenture=0;indenture<fErrorIndent;indenture++) indent = indent + "    ";
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc);
    		if (fErrorIndent != 0) new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
            break;

    case	LOG_INFO:
            if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc);
            break;

    case	LOG_NOTE:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc);
            break;

    case	LOG_ERROR:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc + " --> " + msg);
    		else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc);
            break;

    }
for (int i=0;i<new_strings->Count;i++) fErrorStrings->Add(new_strings->Strings[i]);
new_strings->Clear();
delete new_strings;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg )
{
Log(LOG_ERROR,functiondesc,"Err # " + IntToStr(errorvalue) + " " + ErrorMessage(errorvalue));
Log(LOG_ERROR,functiondesc,"Err : " + errormsg);
}
//---------------------------------------------------------------------------
System::AnsiString	__fastcall	TDirectDraw::ErrorMessage(int errorvalue)
{
System::AnsiString	errormessage = "";
switch (errorvalue)
	{
    // common directx results
    case DD_OK:						errormessage = "DD_OK, DC retrieved successfully"; break;
	case DDERR_GENERIC:             errormessage = "DDERR_GENERIC, generic error"; break;
	case DDERR_INVALIDOBJECT:		errormessage = "DDERR_INVALIDOBJECT, Invalid object"; break;
	case DDERR_INVALIDPARAMS:		errormessage = "DDERR_INVALIDPARAMS, Invalid params"; break;
	case DDERR_INVALIDSURFACETYPE:	errormessage = "DDERR_INVALIDSURFACETYPE, Invalid surface type"; break;
	case DDERR_UNSUPPORTED:			errormessage = "DDERR_UNSUPPORTED, Unsupported"; break;
	case DDERR_SURFACELOST:			errormessage = "DDERR_SURFACELOST, Surface is lost"; break;
	case DDERR_SURFACEBUSY:			errormessage = "DDERR_SURFACEBUSY, Surface is busy"; break;
	case DDERR_WASSTILLDRAWING:		errormessage = "DDERR_WASSTILLDRAWING, Was still drawing"; break;
	// flip results
	case DDERR_NOFLIPHW:			errormessage = "DDERR_NOFLIPHW, No flip hardware"; break;
	case DDERR_NOTFLIPPABLE:		errormessage = "DDERR_NOTFLIPPABLE, Surface is not flippable"; break;
    // GetDC results
	case DDERR_DCALREADYCREATED:	errormessage = "DDERR_DCALREADYCREATED, DC was already created"; break;
    default:						errormessage = "UNKNOWN ERROR, the error generated is unknown";
    };
return errormessage;
}
//---------------------------------------------------------------------------
/*void	__fastcall	TDirectDraw::Clear(TColor color)
{
Log(LOG_IN,"TDirectDraw::Clear()","color="+IntToStr(color));
if (fBackBufferUsedLast)
	{
	Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the back buffers memory");
	memset( fBackBufferLockingDesc.lpSurface, color, fWidth * fHeight );
	Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the back buffer");
	Flip();
    }
else
    if (fDrawingOnPrimarySurface)
		{
		Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the primary surfaces memory");
		memset( fPrimarySurfaceLockingDesc.lpSurface, color, fWidth * fHeight );
		Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the primary surface");
        Flip();
	    }
Log(LOG_OUT,"TDirectDraw::Clear()","");
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::LockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
Log(LOG_IN,"TDirectDraw::LockSurface()","surface=" + IntToStr((int)surface));

// setup the surface description we use for locking / unlocking surfaces
Log(LOG_NOTE,"TDirectDraw::Constructor()","setting up the locking / unlocking surface descriptions");
memset(&fPrimarySurfaceLockingDesc,0,sizeof(fPrimarySurfaceLockingDesc));
fPrimarySurfaceLockingDesc.dwSize = sizeof(fPrimarySurfaceLockingDesc);
memset(&fBackBufferLockingDesc,0,sizeof(fBackBufferLockingDesc));
fBackBufferLockingDesc.dwSize = sizeof(fBackBufferLockingDesc);

if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::LockSurface()","surface cannot be locked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (!fBackBufferLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the back buffer");
	        fErrorValue = surface->Lock(NULL,&fBackBufferLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the back buffer could not be locked");
		        return false;
	            }
            fBackBufferLocked = true;
			fBackBufferUsedLast = true;
            fPrimarySurfaceUsedLast = false;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the back buffer locked successfully");
        	}
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the back buffer is already locked");
	    }
    else
    	if (!fPrimarySurfaceLocked)
	    	{
    	    Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the primary surface");
	        fErrorValue = surface->Lock(NULL,&fPrimarySurfaceLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the primary surface could not be locked");
		        return false;
	            }
            fPrimarySurfaceLocked = true;
            fBackBufferUsedLast = false;
			fPrimarySurfaceUsedLast = true;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the primary surface locked successfully");
	        }
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the primary surface is already locked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::LockSurface()","");
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::UnLockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
Log(LOG_IN,"TDirectDraw::UnLockSurface()","surface=" + IntToStr((int)surface));
if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","surface cannot be unlocked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (fBackBufferLocked)
        	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the back buffer");
			surface->Unlock(fBackBufferLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the back buffer could not be unlocked");
		        return false;
	            }
            fBackBufferLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the back buffer unlocked successfully");
            }
        else
		    Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","the back buffer was already unlocked");
	    }
    else
    	if (fPrimarySurfaceLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the primary surface");
			fErrorValue = surface->Unlock(fPrimarySurfaceLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the primary surface could not be unlocked");
		        return false;
	            }
	        fPrimarySurfaceLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface unlocked successfully");
	        }
        else
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface was already unlocked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::UnLockSurface()","");
return false;
}
*/


/*
============================================================================================
============================================================================================
============================================================================================
============================================================================================
============================================================================================
============================================================================================
MOUSE CURSOR STUFF
============================================================================================
============================================================================================
============================================================================================
============================================================================================
============================================================================================
============================================================================================
	if (!MOUSE_INSTALLED) return;
	if (mouse_IsHidden()) return;

	// update only if the cursor has moved!
	if (HasMoved)
		{
//		bool save_disabled = MOUSE_ISR_DISABLED;
//		MOUSE_ISR_DISABLED = true;

		// undraw the cursor at the last drawn coords

		if (!MOUSE_FIRST_TIME_DRAW) mouse_UndrawCursor();

		// draw the cursor at the new x,y coords

		mouse_DrawCursor( SAVE_UNDERNEATH_CURSOR );

		// reset flag so that we don't keep drawing cursor

		HasMoved = false;
		MOUSE_FIRST_TIME_DRAW = false;

//		MOUSE_ISR_DISABLED = save_disabled;
		}
*/
