//---------------------------------------------------------------------------
//
//	TDirectDraw - Borland C++ Builder / DirectDraw implementation
//
//	(c) 1997 Darren J. Dwyer  - All Rights Reserved
//
// Disclaimer:
// All source code, object files, documentation files, etc. included
// with this package are copyright. Any use contrary to the attached
// file LICENCE.DOC are deemed to be breaches of copyright.
//
//---------------------------------------------------------------------------
#ifndef TDirectDrawH
#define TDirectDrawH
//---------------------------------------------------------------------------
#include <vcl\SysUtils.hpp>
#include <vcl\Controls.hpp>
#include <vcl\Classes.hpp>
#include <vcl\Forms.hpp>
//---------------------------------------------------------------------------
#include <d3drmwin.h>
#include <direct.h>
//---------------------------------------------------------------------------
// Various error levels used by the Log() function. See below for more info.
//---------------------------------------------------------------------------
#define	LOG_IN		0
#define	LOG_OUT		1
#define	LOG_INFO	2
#define	LOG_NOTE	3
#define	LOG_ERROR	4
//---------------------------------------------------------------------------
enum  TDirectDrawMode { dm640x480x8, dm640x480x16,
						dm800x600x8, dm800x600x16,
						dm1024x768x8, dm1024x768x16 };
//---------------------------------------------------------------------------
class TDirectDraw : public TComponent	// note: 	don't _ever_ put component
										//			onto a TDirectDrawForm
{
    //----------------------------------------------------------------------------------
	// General Notes:
    //
    // (1)	This object "TDirectDraw" is designed primarily to be used with the
    // 		"TDirectDrawForm" object written by the same author. If this object
    //		is used separately, note that you should create only ONE (1) instance
    //		of this object. Users of the "TDirectDrawForm" object should never need
    //		to create an instance of this object as this is all handled within the
    //		"TDirectDrawForm" object.
    //
    // (2)	DirectDraw consists of a number of drawing surfaces that can be swapped,
    // 		drawn on and otherwise manipulated. Of these surfaces, the currently
    //		visible surface is called the primary surface. Attached to this surface
    //		are one or more hidden surfaces, stored in a linked-list arrangement, the
    //		first being called the back buffer. Only the primary surface and the back
    //      buffer can be manipulated at any one time, although they can be rotated
    //		so that you can draw on any or all of the available surfaces. See: Flip()
    //
    //----------------------------------------------------------------------------------

private:
	TDirectDrawMode			fDrawMode;
    int						fBackSurfaceCount;

	void	__fastcall	SetDrawMode( TDirectDrawMode drawmode )	{ fDrawMode = drawmode; }
    void	__fastcall	SetBackSurfaceCount( int surfacecount ) { fBackSurfaceCount = surfacecount; }

__published:

	__property	TDirectDrawMode		DrawMode = {read=fDrawMode,write=SetDrawMode,default=dm800x600x16};
    __property	int					BackSurfaceCount = {read=fBackSurfaceCount,write=fBackSurfaceCount,default=1};

public:
						__fastcall 	TDirectDraw(TComponent* Owner);
    virtual				__fastcall 	~TDirectDraw();

    //----------------------------------------------------------------------------------
    // Func:	GetPrimarySurfaceDC(), GetBackBufferDC()
    // Desc:	Use these functions to return a HDC (device context) that can be used by
    //			various gdi functions.
    // Returns:	If a valid DC is available, the functions return a HDC
    //			If no DC is available, these functions return NULL
	//
    // Note: 	If called and the DC is valid (ie. not NULL), you must call the
    //			appropriate ReleasePrimarySurfaceDC() or ReleaseBackBufferDC() functions
    //			before returning program control to windows.
    //----------------------------------------------------------------------------------
    HDC					__fastcall	GetBackBufferDC();
    HDC					__fastcall	GetPrimarySurfaceDC();

    //----------------------------------------------------------------------------------
    // Func:	ReleasePrimarySurfaceDC(), ReleaseBackBufferDC()
    // Desc:	Use these functions to release resources that were allocated when either
    //			GetPrimarySurfaceDC() or GetBackBufferDC() functions have been called.
	// Returns:	Upon success, these functions return TRUE, otherwise they return FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	ReleaseBackBufferDC();
    bool				__fastcall	ReleasePrimarySurfaceDC();

    //----------------------------------------------------------------------------------
    // Func:	FlipToGDI()
    // Desc:	This function will continue to flip surfaces until the surface the GDI
    //			uses becomes the primary surface.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	FlipToGDI();

    //----------------------------------------------------------------------------------
    // Func:	Flip()
    // Desc:    This function will rotate all current DirectDraw surfaces so that the
    //			back buffer will become the primary surface, the next back buffer will
    //			become the back buffer (where there are multiple back buffers) and the
    //			primary surface will become the last back buffer.
    // Params:	wait - 	if TRUE, this function will wait until the surfaces are
    //					available before flipping. if FALSE, the flip may not occur.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
	bool				__fastcall	Flip(bool wait = true);

    //----------------------------------------------------------------------------------
    // Func:	FlipPrimaryToBack()
    // Desc:    This function will copy the contents of the primary surface to the
    //			back buffer using accelerated functions if available.
    // Params:	wait - 	if TRUE, this function will wait until the surfaces are
    //					available before flipping. if FALSE, the flip may not occur.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	FlipPrimaryToBack(bool wait = true);

    //----------------------------------------------------------------------------------
    // Func:	FlipBackToPrimary()
    // Desc:    This function will copy the contents of the back buffer to the
    //			primary surface using accelerated functions if available.
    // Params:	wait - 	if TRUE, this function will wait until the surfaces are
    //					available before flipping. if FALSE, the flip may not occur.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	FlipBackToPrimary(bool wait = true);

    //----------------------------------------------------------------------------------
    // Func:	FlipPrimaryRectToBack()
    // Desc:    This function will copy a rectangular section of the primary surface to
    //			the back buffer using accelerated functions if available.
    // Params:	left,top,right,bottom -	the coords of the rectangle to copy
    //			wait - 	if TRUE, this function will wait until the surfaces are
    //					available before flipping. if FALSE, the flip may not occur.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	FlipPrimaryRectToBack(int left,int top,int width,int height, bool wait = true);

    //----------------------------------------------------------------------------------
    // Func:	FlipBackRectToPrimary()
    // Desc:    This function will copy a rectangular section of the back buffer to
    //			the primary surface using accelerated functions if available.
    // Params:	left,top,right,bottom -	the coords of the rectangle to copy
    //        	wait - 	if TRUE, this function will wait until the surfaces are
    //					available before flipping. if FALSE, the flip may not occur.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	FlipBackRectToPrimary(int left,int top,int width,int height, bool wait = true);

    //----------------------------------------------------------------------------------
	// Func:	PutBackBufferImage(), PutPrimarySurfaceImage()
    // Desc:	These functions will draw the TImage to the appropriate surface using
    //			the supplied copymode.
    // Params:	left,top,width,height - these are the coordinates where the image will be drawn
    //			image - the image to be drawn
    //			copymode - the method used to draw the image
    // Returns:	Upon success, these functions return TRUE, otherwise they return FALSE
    //----------------------------------------------------------------------------------
    bool				__fastcall	PutBackBufferImage(int left,int top,int width,int height,TImage* image,long copymode = cmSrcCopy);
    bool				__fastcall	PutPrimarySurfaceImage(int left,int top,int width,int height,TImage* image,long copymode = cmSrcCopy);

    //----------------------------------------------------------------------------------
	// Func:	GetBackBufferImage(), GetPrimarySurfaceImage()
    // Desc:	These functions will create a new TImage and will fill in the image with
    //			pixels from the appropriate surface.
    // Params:	x,y		- these are the top, left coordinates where the image will be drawn
    //			image	- the image to be drawn
    //			copymode- the method used to draw the image
    // Returns:	Upon success, these functions return the finished image, otherwise they
    //			return NULL.
    //
    // Note:	Since the image is newly created, you should delete it when finished.
    //----------------------------------------------------------------------------------
    TImage*				__fastcall	GetBackBufferImage(int left,int top,int width,int height);
    TImage*				__fastcall	GetPrimarySurfaceImage(int left,int top,int width,int height);

    //----------------------------------------------------------------------------------
	// Func:	SetDisplayMode()
    // Desc:	This function will attempt to set the current display mode to the
    //			resolution specified.
    // Returns:	Upon success, this function returns TRUE, otherwise it returns FALSE
    //----------------------------------------------------------------------------------
	bool				__fastcall	SetDisplayMode(int width,int height,int bitsperpixel);

    //----------------------------------------------------------------------------------
	// Func:	PutImage(), GetImage()
    // Desc:	These functions will draw / retrieve a TImage to / from the specified DC.
    // Returns:	Upon success, these functions return TRUE, otherwise they return FALSE
    // Note:	Really meant to be used internally, but can be used otherwise
    //----------------------------------------------------------------------------------
    bool				__fastcall	PutImage(HDC hdc,int left,int top,int width,int height,TImage* image,long copymode);
    TImage*				__fastcall	GetImage(HDC hdc,int left,int top,int width,int height);

    //----------------------------------------------------------------------------------
    // Prop:	DirectDraw - returns a pointer to the DirectDraw object used in this unit.
    // Prop:	PrimarySurface - returns a pointer to the primary surface.
	// Prop:	BackBuffer - returns a pointer to the back buffer.
    //
    // Notes:	Use these properties to access the actual DirectDraw objects in the
    //			case that you want to perform more advanced DirectDraw functions.
    // Result:	If the result is NULL, the object has not been created or for some
    //			other reason does not exist.
    //----------------------------------------------------------------------------------
	__property	LPDIRECTDRAW			DirectDraw = {read=fDirectDraw};
    __property	LPDIRECTDRAWSURFACE		PrimarySurface = {read=fPrimarySurface};
    __property	LPDIRECTDRAWSURFACE		BackBuffer = {read=fBackBuffer};

    //----------------------------------------------------------------------------------
	// Func:	Log(), Error(), ErrorMessage()
    // Desc:	This function can be used to "Log" an message for later retrieval using
    // 			the ErrorStrings property.
    // Params:	msgclass - one of LOG_IN, LOG_OUT, LOG_INFO, LOG_NOTE, LOG_ERROR
    //			functiondesc - generally meant to be a function reference.
    //			msg - the message to display
    //			errorvalue - the result from calling a DirectDraw function.
    // Example:	void MyFunc()
    //			{
    //			DirectDraw->Log( LOG_IN, Name+"::MyFunc()", "Entering func");
    //			.
    //			.
    //				DirectDraw->Log( LOG_INFO, Name+"::MyFunc()", "Middle of func");
    //			.
    //			.
    //          DirectDraw->Log( LOG_OUT, Name+"::MyFunc()", "Exitting func");
    //          }
    // Note:	These functions require that DirectDraw->Debugging == true;
    //----------------------------------------------------------------------------------
    void				__fastcall	Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg );
    void				__fastcall	Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg );
    System::AnsiString	__fastcall	ErrorMessage(int errorvalue);

    //----------------------------------------------------------------------------------
	// Prop:	Debugging	 - if FALSE, the Log() function does not work.
    // Prop:	ErrorValue 	 - the last result from any DirectDraw function call
    // Prop:	ErrorCount 	 - the number of messages that have been "logged" since
    //							startup or since setting DirectDraw->ErrorCount = n;
    // Prop:	ErrorStrings - a list of all messages logged since startup or since
    //							DirectDraw->ErrorStrings->Clear();
    //----------------------------------------------------------------------------------
	__property 	bool				Debugging = {read=fDebugging,write=fDebugging,default=false};
    __property	HRESULT				ErrorValue = {read=fErrorValue};
    __property	int					ErrorCount = {read=fErrorCount,write=fErrorCount};
    __property	TStringList*		ErrorStrings = {read=fErrorStrings};

private:
    bool							fDebugging;
    HRESULT							fErrorValue;
    int								fErrorCount;
    TStringList*					fErrorStrings;
    int								fErrorIndent;

	LPDIRECTDRAW					fDirectDraw;
    LPDIRECTDRAWSURFACE				fPrimarySurface;
    LPDIRECTDRAWSURFACE				fBackBuffer;

    HDC								fBackBufferDC;
    HDC								fPrimarySurfaceDC;
    int								fWidth;
    int								fHeight;
    int								fBitsPerPixel;
    int								fBackBufferCount;

    POINT							fMouseCursorPos;
    POINT							fMouseCursorLastPos;
    Graphics::TBitmap*				fMouseCursorBkgnd;
    bool							fMouseFirstTimeDraw;

    // functions used internally
    bool				__fastcall	CreateDirectDraw();
    bool				__fastcall	ReleaseDirectDraw();
    bool				__fastcall	SetExclusiveMode();
    bool				__fastcall	SetNormalMode();
    bool				__fastcall	RestoreDisplayMode();
	bool				__fastcall	CreateSurfaces();
    bool				__fastcall	ReleaseSurfaces();

protected:
    bool				__fastcall	DrawMouseCursor(HDC hdc);
};
//---------------------------------------------------------------------------
extern  TDirectDraw*	DirectDraw;
//---------------------------------------------------------------------------
#endif

